from __future__ import absolute_import, division, print_function

import six
import datetime
import json
from copy import deepcopy

from locker import util, binary_adapter, DEFAULT_TIMEOUT


def _compute_diff(current, previous):
    if isinstance(current, dict):
        previous = previous or {}
        diff = current.copy()
        for key in set(previous.keys()) - set(diff.keys()):
            diff[key] = ""
        return diff
    return current if current is not None else ""


def _serialize_list(array, previous):
    array = array or []
    previous = previous or []
    params = {}

    for i, v in enumerate(array):
        previous_item = previous[i] if len(previous) > i else None
        if hasattr(v, "serialize"):
            params[str(i)] = v.serialize(previous_item)
        else:
            params[str(i)] = _compute_diff(v, previous_item)

    return params


class LSObject(dict):
    class ReprJSONEncoder(json.JSONEncoder):
        def default(self, obj):
            if isinstance(obj, datetime.datetime):
                return util.encode_datetime(obj)
            return super(LSObject.ReprJSONEncoder, self).default(obj)

    def __init__(self, id=None, access_key_id=None, secret_access_key=None, api_base=None, api_version=None,
                 last_response=None, **params):
        super().__init__()

        self._unsaved_values = set()
        self._transient_values = set()
        self._last_response = last_response

        self._retrieve_params = params
        self._previous = None

        object.__setattr__(self, "access_key_id", access_key_id)
        object.__setattr__(self, "secret_access_key", secret_access_key)
        object.__setattr__(self, "api_base", api_base)
        object.__setattr__(self, "api_version", api_version)

        if id:
            self["id"] = id

    @property
    def last_response(self):
        return self._last_response

    def update(self, update_dict):
        for k in update_dict:
            self._unsaved_values.add(k)

        return super().update(update_dict)

    def __setattr__(self, k, v):
        if k[0] == "_" or k in self.__dict__:
            return super().__setattr__(k, v)
        self[k] = v
        return None

    def __getattr__(self, k):
        if k[0] == "_":
            raise AttributeError(k)
        try:
            return self[k]
        except KeyError as err:
            raise AttributeError(*err.args)

    def __delattr__(self, k):
        if k[0] == "_" or k in self.__dict__:
            return super().__delattr__(k)
        else:
            del self[k]

    def __setitem__(self, k, v):
        if v == "":
            raise ValueError(
                "You cannot set %s to an empty string on this object. "
                "The empty string is treated specially in our requests. "
                "If you'd like to delete the property using the save() method on this object, you may set %s.%s=None. "
                "Alternatively, you can pass %s='' to delete the property when "
                "using a resource method such as modify()." % (k, str(self), k, k)
            )
        # Allows for unpickling in Python 3.x
        if not hasattr(self, "_unsaved_values"):
            self._unsaved_values = set()
        self._unsaved_values.add(k)
        super().__setitem__(k, v)

    def __getitem__(self, k):
        try:
            return super().__getitem__(k)
        except KeyError as err:
            if k in self._transient_values:
                raise KeyError(
                    "%r.  HINT: The %r attribute was set in the past."
                    "It was then wiped when refreshing the object with "
                    "the result returned by Locker's API, probably as a "
                    "result of a save().  The attributes currently "
                    "available on this object are: %s"
                    % (k, k, ", ".join(list(self.keys())))
                )
            else:
                raise err

    def __delitem__(self, k):
        super().__delitem__(k)

        # Allows for unpickling in Python 3.x
        if hasattr(self, "_unsaved_values") and k in self._unsaved_values:
            self._unsaved_values.remove(k)

    # Custom unpickling method that uses `update` to update the dictionary
    # without calling __setitem__, which would fail if any value is an empty
    # string
    def __setstate__(self, state):
        self.update(state)

    # Custom pickling method to ensure the instance is pickled as a custom
    # class and not as a dict, otherwise __setstate__ would not be called when
    # unpickling.
    def __reduce__(self):
        reduce_value = (
            type(self),  # callable
            (   # args
                self.get("id", None),
                self.access_key,
                self.api_base,
                self.api_version,
            ),
            dict(self),  # state
        )
        return reduce_value

    @classmethod
    def construct_from(cls, values, access_key_id, secret_access_key, api_base=None, api_version=None,
                       last_response=None):
        instance = cls(
            values.get("id"),
            access_key_id=access_key_id,
            secret_access_key=secret_access_key,
            api_base=api_base,
            api_version=api_version,
            last_response=last_response,
        )
        instance.refresh_from(
            values,
            access_key_id=access_key_id,
            secret_access_key=secret_access_key,
            api_base=api_base,
            api_version=api_version,
            last_response=last_response,
        )
        return instance

    def refresh_from(self, values, access_key_id=None, secret_access_key=None, partial=False,
                     api_base=None, api_version=None, last_response=None):
        self.access_key_id = access_key_id or getattr(values, "access_key_id", None)
        self.secret_access_key = secret_access_key or getattr(values, "secret_access_key", None)
        self.api_base = api_base or getattr(values, "api_base", None)
        self.api_version = api_version or getattr(values, "api_version", None)
        self._last_response = last_response or getattr(values, "_last_response", None)

        # Wipe old state before setting new.  This is useful for e.g.
        # updating a customer, where there is no persistent card
        # parameter.  Mark those values which don't persist as transient
        if partial:
            self._unsaved_values = self._unsaved_values - set(values)
        else:
            removed = set(self.keys()) - set(values)
            self._transient_values = self._transient_values | removed
            self._unsaved_values = set()
            self.clear()

        self._transient_values = self._transient_values - set(values)

        for k, v in six.iteritems(values):
            super().__setitem__(
                k,
                util.convert_to_ls_object(v, access_key_id, secret_access_key, api_base, api_version),
            )

        self._previous = values

    def call(self, cli, asjson=True, timeout=DEFAULT_TIMEOUT, output_format="json", params=None):
        return LSObject._call(
            self, cli, asjson=asjson, timeout=timeout, output_format=output_format, params=params
        )

    # The 'cli_' is suffixed with an underscore to avoid conflicting with request parameters in `params`
    def _call(self, cli_, access_key_id=None, secret_access_key=None, api_base=None, api_version=None,
              asjson=True, timeout=DEFAULT_TIMEOUT, output_format="json", params=None):
        params = None if params is None else params.copy()

        access_key_id = util.read_special_variable(params, "access_key_id", access_key_id)
        secret_access_key = util.read_special_variable(params, "secret_access_key", secret_access_key)
        api_base = util.read_special_variable(params, "api_base", api_base)
        api_version = util.read_special_variable(params, "api_version", api_version)
        headers = util.read_special_variable(params, "headers", None)
        skip_cli_lines = util.read_special_variable(params, "skip_cli_lines", None)
        logger = util.read_special_variable(params, "logger", None) if params else None
        resttime = util.read_special_variable(params, "resttime", None)
        fetch = util.read_special_variable(params, "fetch", None)

        access_key_id = access_key_id or self.access_key_id
        secret_access_key = secret_access_key or self.secret_access_key
        api_base = api_base or self.api_base
        api_version = api_version or self.api_version
        params = params or self._retrieve_params

        binary_executor = binary_adapter.BinaryAdapter(
            access_key_id=access_key_id,
            secret_access_key=secret_access_key,
            api_base=api_base,
            api_version=api_version,
            headers=headers,
            logger=logger,
            resttime=resttime,
            fetch=fetch,
        )
        res_data = binary_executor.call(
            cli=cli_, params=params, asjson=asjson, timeout=timeout, skip_cli_lines=skip_cli_lines,
            output_format=output_format
        )
        return util.convert_to_ls_object(
            res_data, access_key_id, secret_access_key, api_base, api_version, params
        )

    def __repr__(self):
        ident_parts = [type(self).__name__]

        if isinstance(self.get("object"), six.string_types):
            ident_parts.append(self.get("object"))

        if isinstance(self.get("id"), six.string_types):
            ident_parts.append("id=%s" % (self.get("id"),))

        unicode_repr = "<%s at %s> JSON: %s" % (
            " ".join(ident_parts),
            hex(id(self)),
            str(self),
        )
        return unicode_repr

    def __str__(self):
        return json.dumps(
            self.to_dict_recursive(),
            sort_keys=True,
            indent=2,
            cls=self.ReprJSONEncoder,
        )

    def to_dict(self):
        return dict(self)

    def to_dict_recursive(self):
        def maybe_to_dict_recursive(value):
            if value is None:
                return None
            elif isinstance(value, LSObject):
                return value.to_dict_recursive()
            else:
                return value

        return {
            key: list(map(maybe_to_dict_recursive, value))
            if isinstance(value, list)
            else maybe_to_dict_recursive(value)
            for key, value in six.iteritems(dict(self))
        }

    @property
    def locker_id(self):
        return self.id

    def serialize(self, previous):
        params = {}
        unsaved_keys = self._unsaved_values or set()
        previous = previous or self._previous or {}

        from locker.ls_resources.abstract.api_resource import APIResource
        for k, v in six.iteritems(self):
            if k == "id" or (isinstance(k, str) and k.startswith("_")):
                continue
            elif isinstance(v, APIResource):
                continue
            elif hasattr(v, "serialize"):
                child = v.serialize(previous.get(k, None))
                if child != {}:
                    params[k] = child
            elif k in unsaved_keys:
                params[k] = _compute_diff(v, previous.get(k, None))
            elif k == "additional_owners" and v is not None:
                params[k] = _serialize_list(v, previous.get(k, None))

        return params

        # This class overrides __setitem__ to throw exceptions on inputs that it
        # doesn't like. This can cause problems when we try to copy an object
        # wholesale because some data that's returned from the API may not be valid
        # if it was set to be set manually. Here we override the class' copy
        # arguments so that we can bypass these possible exceptions on __setitem__.

    def __copy__(self):
        copied = LSObject(
            self.get("id"),
            self.access_key,
            api_base=self.api_base,
            api_version=self.api_version,
        )

        copied._retrieve_params = self._retrieve_params

        for k, v in six.iteritems(self):
            # Call parent's __setitem__ to avoid checks that we've added in the
            # overridden version that can throw exceptions.
            super(LSObject, copied).__setitem__(k, v)

        return copied

        # This class overrides __setitem__ to throw exceptions on inputs that it
        # doesn't like. This can cause problems when we try to copy an object
        # wholesale because some data that's returned from the API may not be valid
        # if it was set to be set manually. Here we override the class' copy
        # arguments so that we can bypass these possible exceptions on __setitem__.

    def __deepcopy__(self, memo):
        copied = self.__copy__()
        memo[id(self)] = copied

        for k, v in six.iteritems(self):
            # Call parent's __setitem__ to avoid checks that we've added in the
            # overridden version that can throw exceptions.
            super(LSObject, copied).__setitem__(k, deepcopy(v, memo))

        return copied



import os
import tempfile
import time

DEFAULT_POLL_TIME = .1


class AtomicNameLock(object):

    def __init__(self, name, lock_dir=None, max_lock_age=None):
        """
        This uses a named directory, which is defined by POSIX as an atomic operation.
        :param name: (str) The lock name, Cannot contain directory seperator (like '/')
        :param lock_dir: (str/None) Directory in which to store locks. Defaults to tempdir
        :param max_lock_age: (float/None) Maximum number of seconds lock can be held before it is considered "too old"
        and fair game to be taken.
        You should likely define this as a reasonable number, maybe 4x as long as you think the operation will take,
        so that the lock doesn't get held by a dead process.
        """
        self.name = name
        self.max_lock_age = max_lock_age

        if os.sep in name:
            raise ValueError('Name cannot contain "%s"' % (os.sep,))

        if lock_dir:
            if lock_dir[-1] == os.sep:
                lock_dir = lock_dir[:-1]
                if not lock_dir:
                    raise ValueError('lockDir cannot be ' + os.sep)
        else:
            lock_dir = tempfile.gettempdir()

        self.lock_dir = lock_dir

        if not os.path.isdir(lock_dir):
            raise ValueError('lockDir %s either does not exist or is not a directory.' % (lock_dir,))

        if not os.access(lock_dir, os.W_OK):
            raise ValueError('Cannot write to lock directory: %s' % (lock_dir,))
        self.lock_path = lock_dir + os.sep + name

        self.held = False
        self.acquired_at = None

    def acquire(self, timeout=None):
        """
        Acquire given lock. Can be blocking or nonblocking by providing a timeout.
        Returns "True" if you got the lock, otherwise "False"
        :param timeout: (float/None) Max number of seconds to wait, or None to block until we can acquire it.
        :return: (bool) True if you got the lock, otherwise False.
        """

        if self.held is True:
            # NOTE: Without some type of in-directory marker (like an uuid) we cannot refresh an expired lock accurately
            if os.path.exists(self.lock_path):
                return True
            # Someone removed our lock
            self.held = False

        # If we aren't going to poll at least 5 times, give us a smaller interval
        if timeout:
            if timeout / 5.0 < DEFAULT_POLL_TIME:
                poll_time = timeout / 10.0
            else:
                poll_time = DEFAULT_POLL_TIME

            end_time = time.time() + timeout
            keep_going = lambda: bool(time.time() < end_time)
        else:
            poll_time = DEFAULT_POLL_TIME
            keep_going = lambda: True

        success = False
        while keep_going():
            try:
                os.mkdir(self.lock_path)
                success = True
                break
            except:
                time.sleep(poll_time)
                if self.max_lock_age:
                    if os.path.exists(self.lock_path) and \
                            os.stat(self.lock_path).st_mtime < time.time() - self.max_lock_age:
                        try:
                            os.rmdir(self.lock_path)
                        except:
                            # If we did not remove the lock, someone else is at the same point and contending.
                            # Let them win.
                            time.sleep(poll_time)

        if success is True:
            self.acquired_at = time.time()

        self.held = success
        return success

    def release(self, force_release=False):
        """
        Release the lock.
        :param force_release: (bool) If True, will release the lock even if we don't hold it.
        :return: True if lock is released, otherwise False
        """
        if not self.held:
            if force_release is False:
                return False  # We were not holding the lock
            else:
                self.held = True  # If we have force release set, pretend like we held its

        if not os.path.exists(self.lock_path):
            self.held = False
            self.acquired_at = None
            return True

        if force_release is False:
            # We waited too long and lost the lock
            if self.max_lock_age and time.time() > self.acquired_at + self.max_lock_age:
                self.held = False
                self.acquired_at = None
                return False

        self.acquired_at = None

        try:
            os.rmdir(self.lock_path)
            self.held = False
            return True
        except:
            self.held = False
            return False

    def __checkExpiration(self, mtime=None):
        """
        Check if we have expired
        :param mtime: (int/None) Optional mtime if known, otherwise will be gathered
        :return: rue if we did expire, otherwise False
        """
        if not self.max_lock_age:
            return False

        if mtime is None:
            try:
                mtime = os.stat(self.lock_path).st_mtime
            except FileNotFoundError as e:
                return False

        if mtime < time.time() - self.max_lock_age:
            return True

        return False

    @property
    def is_held(self):
        """
        True if anyone holds the lock, otherwise False.
        :return: If lock is held by anyone
        """
        if not os.path.exists(self.lock_path):
            return False

        try:
            mtime = os.stat(self.lock_path).st_mtime
        except FileNotFoundError as e:
            return False

        if self.__checkExpiration(mtime):
            return False

        return True

    @property
    def has_lock(self):
        """
        Property, returns True if we have the lock, or False if we do not.
        :return: True/False if we have the lock or not.
        """
        # If we don't hold it currently, return False
        if self.held is False:
            return False

        # Otherwise if we think we hold it, but it is not held, we have lost it.
        if not self.is_held:
            self.acquired_at = None
            self.held = False
            return False

        # Check if we expired
        if self.__checkExpiration(self.acquired_at):
            self.acquired_at = None
            self.held = False
            return False

        return True



# -*- coding: utf-8 -*-
# File generated from our OpenAPI spec
from __future__ import absolute_import, division, print_function

from locker import ls_resources


OBJECT_CLASSES = {
    # data structures
    # ls_resources.ListObject.OBJECT_NAME: ls_resources.ListObject,
    # business objects
    ls_resources.Secret.OBJECT_NAME: ls_resources.Secret,
    ls_resources.Environment.OBJECT_NAME: ls_resources.Environment
}



from __future__ import absolute_import, division, print_function

import locker


class LockerError(Exception):
    def __init__(self, message=None,
                 http_body=None, http_status=None, json_body=None, headers=None,
                 code=None,
                 status_code=None,
                 ):
        super().__init__(message)

        if http_body and hasattr(http_body, "decode"):
            try:
                http_body = http_body.decode("utf-8")
            except BaseException:
                http_body = (
                    "<Could not decode body as utf-8. Please report to support@locker.io>"
                )

        self._message = message
        self.http_body = http_body
        self.http_status = http_status
        self.json_body = json_body
        self.headers = headers or {}
        self.code = code
        self.status_code = status_code
        self.request_id = self.headers.get("request-id", None)
        self.error = self.construct_error_object()

    def __str__(self):
        msg = self._message or "<empty message>"
        if self.request_id is not None:
            return u"Request {0}: {1}".format(self.request_id, msg)
        else:
            return msg

    @property
    def user_message(self):
        return self._message

    def __repr__(self):
        return "%s(message=%r, http_status=%r, request_id=%r)" % (
            self.__class__.__name__,
            self._message,
            self.http_status,
            self.request_id,
        )

    def construct_error_object(self):
        if (
            self.json_body is None
            or "error" not in self.json_body
            # or not isinstance(self.json_body["error"], dict)
        ):
            return None

        return locker.ls_resources.error_object.ErrorObject.construct_from(
            self.json_body, locker.access_key
        )


class APIError(LockerError):
    pass


class CliRunError(LockerError):
    """."""

    process = None

    def __init__(
        self,
        message,
        http_body=None,
        http_status=None,
        json_body=None,
        headers=None,
        code="cli_error",
        status_code=None,
        should_retry=False,
    ):
        super(CliRunError, self).__init__(
            message, http_body, http_status, json_body, headers, code, status_code
        )
        self.should_retry = should_retry


class APIConnectionError(LockerError):
    def __init__(
        self,
        message,
        http_body=None,
        http_status=None,
        json_body=None,
        headers=None,
        code=None,
        status_code=None,
        should_retry=False,
    ):
        super(APIConnectionError, self).__init__(
            message, http_body, http_status, json_body, headers, code, status_code
        )
        self.should_retry = should_retry


class APIServerError(LockerError):
    pass


class LockerErrorWithParamCode(LockerError):
    def __init__(
        self,
        message,
        param,
        code=None,
        status_code=None,
        http_body=None,
        http_status=None,
        json_body=None,
        headers=None,
    ):
        super(LockerErrorWithParamCode, self).__init__(
            message, http_body, http_status, json_body, headers, code, status_code
        )
        self.param = param

    def __repr__(self):
        return ("%s(message=%r, param=%r, code=%r, http_status=%r, request_id=%r)" % (
            self.__class__.__name__,
            self._message,
            self.param,
            self.code,
            self.http_status,
            self.request_id,
        ))


class IdempotencyError(LockerError):
    pass


class InvalidRequestError(LockerErrorWithParamCode):
    def __init__(
        self,
        message,
        param,
        code=None,
        status_code=None,
        http_body=None,
        http_status=None,
        json_body=None,
        headers=None,
    ):
        super(InvalidRequestError, self).__init__(
            message, http_body, http_status, json_body, headers, code, status_code
        )
        self.param = param


class AuthenticationError(LockerError):
    pass


class PermissionDeniedError(LockerError):
    pass


class ResourceNotFoundError(LockerError):
    pass


class RateLimitError(LockerError):
    pass





import os
import re
import sys
import six
import traceback
import logging


LOCKER_LOG = os.environ.get("LOCKER_LOG")

logger = logging.getLogger("locker")


def logfmt(props):
    def fmt(key, val):
        # Handle case where val is a bytes or bytesarray
        if hasattr(val, "decode"):
            val = val.decode("utf-8")
        # Check if val is already a string to avoid re-encoding into
        # ascii. Since the code is sent through 2to3, we can't just
        # use unicode(val, encoding='utf8') since it will be
        # translated incorrectly.
        if not isinstance(val, six.string_types):
            val = six.text_type(val)
        if re.search(r"\s", val):
            val = repr(val)
        # key should already be a string
        if re.search(r"\s", key):
            key = repr(key)
        return u"{key}={val}".format(key=key, val=val)

    return u" ".join([fmt(key, val) for key, val in sorted(props.items())])


def _console_log_level(log_level):
    if log_level in ["debug", "info", "warning", "error"]:
        return log_level
    elif LOCKER_LOG in ["debug", "info", "warning", "error"]:
        return LOCKER_LOG
    else:
        return None


def log_debug(message, **params):
    msg = logfmt(dict(message=message, **params))
    # if _console_log_level() == "debug":
    #     print(msg, file=sys.stderr)
    logger.debug(msg)


def log_info(message, **params):
    msg = logfmt(dict(message=message, **params))
    # if _console_log_level() in ["debug", "info"]:
    #     print(msg, file=sys.stderr)
    logger.info(msg)


def log_warning(message, **params):
    msg = logfmt(dict(message=message, **params))
    # if _console_log_level() in ["debug", "info"]:
    #     print(msg, file=sys.stderr)
    logger.warning(msg)


def log_error(message, **params):
    if not message:
        tb = traceback.format_exc()
        message = 'Something was wrong' if tb is None else tb
    msg = logfmt(dict(message=message, **params))
    logger.error(msg)



from __future__ import absolute_import, division, print_function

from locker.ls_resources.abstract.api_resource import APIResource


class DetailableAPIResource(APIResource):
    @classmethod
    def retrieve(cls, id, access_key_id=None, secret_access_key=None, **params):
        return super().retrieve(id, access_key_id, secret_access_key, **params)



from __future__ import absolute_import, division, print_function

import datetime
import calendar
from collections import OrderedDict
import six
import time

import locker
from locker.ls_response import LSResponse
from locker.ls_object import LSObject


def get_object_classes():
    # This is here to avoid a circular dependency
    from locker.object_classes import OBJECT_CLASSES
    return OBJECT_CLASSES


def convert_to_ls_object(
    resp, access_key_id=None, secret_access_key=None, api_base=None, api_version=None, params=None
):
    # This is here to avoid a circular dependency
    # If we get a LSResponse, we'll want to return a
    # LSObject with the last_response field filled out with
    # the raw API response information
    ls_response = None

    if isinstance(resp, locker.ls_response.LSResponse):
        ls_response = resp
        resp = ls_response.data

    if isinstance(resp, list):
        return [
            convert_to_ls_object(i, access_key_id, secret_access_key, api_base, api_version) for i in resp
        ]
    elif isinstance(resp, dict) and not isinstance(
        resp, locker.ls_object.LSObject
    ):
        resp = resp.copy()
        klass_name = resp.get("object")
        if isinstance(klass_name, six.string_types):
            klass = get_object_classes().get(
                klass_name, locker.ls_object.LSObject
            )
        else:
            klass = locker.ls_object.LSObject

        obj = klass.construct_from(
            resp,
            access_key_id=access_key_id,
            secret_access_key=secret_access_key,
            api_base=api_base,
            api_version=api_version,
            last_response=ls_response,
        )

        # We only need to update _retrieve_params when special params were
        # actually passed. Otherwise, leave it as is as the list / search result
        # constructors will instantiate their own params.
        if (
            params is not None
            and hasattr(obj, "object")
            and ((obj.object == "list") or (obj.object == "search_result"))
        ):
            obj._retrieve_params = params
        return obj
    else:
        return resp


def read_special_variable(params, key_name, default_value):
    value = default_value
    params_value = None

    if params is not None and key_name in params:
        params_value = params[key_name]
        del params[key_name]
    if value is None:
        value = params_value
    return value


def encode_datetime(dttime):
    if dttime.tzinfo and dttime.tzinfo.utcoffset(dttime) is not None:
        utc_timestamp = calendar.timegm(dttime.utctimetuple())
    else:
        utc_timestamp = time.mktime(dttime.timetuple())

    return int(utc_timestamp)


def _encode_datetime(dttime):
    if dttime.tzinfo and dttime.tzinfo.utcoffset(dttime) is not None:
        utc_timestamp = calendar.timegm(dttime.utctimetuple())
    else:
        utc_timestamp = time.mktime(dttime.timetuple())

    return int(utc_timestamp)


def _encode_nested_dict(key, data, fmt="%s[%s]"):
    d = OrderedDict()
    for sub_key, sub_value in six.iteritems(data):
        d[fmt % (key, sub_key)] = sub_value
    return d


def api_encode(data):
    for key, value in six.iteritems(data):
        if value is None:
            continue
        elif hasattr(value, "locker_id"):
            yield key, value.stripe_id
        elif isinstance(value, list) or isinstance(value, tuple):
            for i, sv in enumerate(value):
                if isinstance(sv, dict):
                    subdict = _encode_nested_dict("%s[%d]" % (key, i), sv)
                    for k, v in api_encode(subdict):
                        yield k, v
                else:
                    yield "%s[%d]" % (key, i), sv
        elif isinstance(value, dict):
            subdict = _encode_nested_dict(key, value)
            for sub_key, sub_value in api_encode(subdict):
                yield sub_key, sub_value
        elif isinstance(value, datetime.datetime):
            yield key, _encode_datetime(value)
        else:
            yield key, value


def merge_dicts(x, y):
    z = x.copy()
    z.update(y)
    return z



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function


from locker.ls_resources.error_object import ErrorObject

from locker.ls_resources.environment import Environment
from locker.ls_resources.secret import Secret



from locker.logger import Logger


class TestLogger(object):
    message = "This is the test log message"

    def test_debug(self):
        logger = Logger(log_level="debug")
        logger.debug(msg=self.message)

    def test_info(self):
        logger = Logger(log_level="info")
        logger.info(msg=self.message)

    def test_warning(self):
        logger = Logger(log_level="warning")
        logger.warning(msg=self.message)

    def test_error(self):
        logger = Logger(log_level="error")
        logger.error(trace=self.message)



from __future__ import absolute_import, division, print_function

from six.moves.urllib.parse import quote_plus

from locker.ls_resources.abstract.api_resource import APIResource


class UpdateableAPIResource(APIResource):
    @classmethod
    def modify(cls, **params):
        name = params.get("sid") or params.get("name")
        cli = [f"{cls.class_cli()}", "update", "--id", f"{quote_plus(name)}"]
        return cls._static_call(cli, params=params)



from __future__ import absolute_import, division, print_function

from locker.util import merge_dicts
from locker.ls_object import LSObject


class ErrorObject(LSObject):
    def refresh_from(self, values, access_key_id=None, secret_access_key=None, partial=False,
                     api_base=None, api_version=None, last_response=None):
        # Unlike most other API resources, the API will omit attributes in error objects when they have a null value.
        # We manually set default values here to facilitate generic error handling.
        # TODO: Add doc_url, etc... in the ErrorObject
        # {
        #     "code": None,
        #     "decline_code": None,
        #     "message": None,
        #     "param": None,
        #     "payment_intent": None,
        #     "payment_method": None,
        #     "setup_intent": None,
        #     "source": None,
        #     "type": None,
        # }
        values = merge_dicts(
            {
                "doc_url": None,
            },
            values,
        )
        return super(ErrorObject, self).refresh_from(
            values,
            access_key_id,
            secret_access_key,
            partial,
            api_base,
            api_version,
            last_response,
        )



import os
import pytest
from dotenv import load_dotenv

import locker.error
from locker.client import Locker
from locker.ls_resources import Secret, Environment


class TestClient(object):
    @staticmethod
    def init_client():
        load_dotenv()
        access_key_id = os.getenv("ACCESS_KEY_ID")
        secret_access_key = os.getenv("SECRET_ACCESS_KEY")
        client = Locker(
            access_key_id=access_key_id,
            secret_access_key=secret_access_key,
        )
        return client

    def test_init_client(self):
        client = self.init_client()
        access_key_id = os.getenv("ACCESS_KEY_ID")
        secret_access_key = os.getenv("SECRET_ACCESS_KEY")
        assert client.access_key_id == access_key_id
        assert client.secret_access_key == secret_access_key

    def test_secrets(self):
        client = self.init_client()
        if not client.access_key_id or not client.secret_access_key:
            return

        # Test create secret
        test_key = "SDK_TEST_CLIENT"
        test_value = "SDK_TEST_VALUE"
        try:
            new_secret = client.create(key=test_key, value=test_value)
        except locker.error.APIConnectionError:
            new_secret = client.retrieve(key=test_key)
            test_value = new_secret.value
        assert new_secret.key == test_key and new_secret.value == test_value, "The creation secret is invalid"

        # Test list secrets
        secrets = client.list()
        for secret in secrets:
            assert isinstance(secret, Secret)
            assert isinstance(secret.key, str)
            assert isinstance(secret.value, str)
            assert isinstance(secret.description, str) or isinstance(secret.description, None)

        # Test get secret
        secret_value = client.get(key=test_key)
        assert secret_value == test_value and isinstance(secret_value, str)
        secret_value = client.get_secret(key=test_key)
        assert secret_value == test_value and isinstance(secret_value, str)

        # Test retrieve secret
        secret = client.retrieve(key=test_key)
        assert hasattr(secret, "id") and hasattr(secret, "key") and hasattr(secret, "value")
        assert hasattr(secret, "secret_hash")
        assert hasattr(secret, "environment_name") and hasattr(secret, "environment_hash") and \
               hasattr(secret, "environment_id")
        assert secret.key == test_key and secret.value == test_value

        # Test modify the secret
        new_test_value = "SDK_TEST_VALUE_UPDATED"
        secret = client.modify(key=test_key, value=new_test_value)
        assert secret.value == new_test_value

    def test_environments(self):
        client = self.init_client()
        if not client.access_key_id or not client.secret_access_key:
            return

        # Test create env
        test_name = "sdk_env"
        test_external_url = "sdk_env.example.com"
        try:
            new_environment = client.create_environment(name=test_name, external_url=test_external_url)
        except locker.error.APIConnectionError:
            new_environment = client.retrieve_environment(name=test_name)
            test_external_url = new_environment.external_url
        assert new_environment.name == test_name and new_environment.external_url == test_external_url

        # Test list environments
        environments = client.list_environments()
        for environment in environments:
            assert isinstance(environment, Environment)
            assert isinstance(environment.name, str)
            assert isinstance(environment.external_url, str) or environment.external_url is None
            assert isinstance(environment.description, str) or environment.description is None

        # Test get env
        env = client.get_environment(name=test_name)
        assert env.external_url == test_external_url and isinstance(env.external_url, str)

        # Test retrieve secret
        env = client.retrieve_environment(name=test_name)
        assert hasattr(env, "id") and hasattr(env, "name") and hasattr(env, "external_url")
        assert hasattr(env, "hash")
        assert env.name == test_name and env.external_url == test_external_url

        # Test modify the secret
        new_test_external_url = "new_sdk_env.example.com"
        secret = client.modify_environment(name=test_name, external_url=new_test_external_url)
        assert secret.external_url == new_test_external_url

    def test_secrets_with_env(self):
        client = self.init_client()
        if not client.access_key_id or not client.secret_access_key:
            return
        # Test create secret with env
        test_env_name = "test"
        test_env_external_url = "test.locker.io"
        try:
            new_environment = client.create_environment(name=test_env_name, external_url=test_env_external_url)
        except locker.error.APIConnectionError:
            new_environment = client.retrieve_environment(name=test_env_name)
            test_env_external_url = new_environment.external_url
        assert new_environment.name == test_env_name and new_environment.external_url == test_env_external_url

        general_key = "GENERAL_SDK_KEY"
        general_value = "GENERAL_SDK_VALUE"
        test_key = "TEST_SDK_KEY"
        test_value = "TEST_SDK_VALUE"
        try:
            general_secret = client.create(key=general_key, value=general_value)
        except locker.error.APIConnectionError:
            general_secret = client.retrieve(key=general_key)
            general_value = general_secret.value
        assert (general_secret.key == general_key and
                general_secret.value == general_value), "The creation general secret is invalid"

        try:
            test_secret = client.create(key=test_key, value=test_value, environment=test_env_name)
        except locker.error.APIConnectionError:
            test_secret = client.retrieve(key=test_key, environment_name=test_env_name)
            test_value = test_secret.value
        assert (test_secret.key == test_key and
                test_secret.value == test_value), "The creation test secret is invalid"

        # Get general and test secret
        test_general_secret = client.get(key=general_key, environment_name=test_env_name)
        assert test_general_secret == general_value
        t_secret = client.get(key=test_key, environment_name=test_env_name)
        assert t_secret == test_value



from __future__ import absolute_import, division, print_function

from six.moves.urllib.parse import quote_plus
from locker import util
from locker.ls_resources.abstract.api_resource import APIResource


class DeletableAPIResource(APIResource):
    @classmethod
    def _cls_delete(cls, sid, **params):
        # TODO: Get delete resource cli
        url = "%s/%s" % (cls.class_url(), quote_plus(sid))
        return cls._static_request("delete", url, params=params)

    # @util.class_method_variant("_cls_delete")
    def delete(self, **params):
        return self._call_and_refresh(
            self.instance_cli(), params=params
        )



from datetime import datetime

from locker import util
from locker.ls_resources.secret import Secret


class TestUtil(object):
    def test_get_object_classes(self):
        assert isinstance(util.get_object_classes(), dict)

    def test_convert_to_ls_object(self):
        resp = {
            "object": "secret",
            "id": "secret-id",
            "data": {
              "key": "MOCK_KEY_SECRET",
              "value": "MOCK_VALUE_SECRET",
              "description": None
            },
            "key": "MOCK_KEY_SECRET",
            "value": "MOCK_VALUE_SECRET",
            "description": None,
            "creation_date": 1686109221.0,
            "revision_date": 1686109221.0,
            "updated_date": 1686109221.0,
            "deleted_date": None,
            "last_use_date": None,
            "project_id": 1,
            "environment_id": None
        }
        obj = util.convert_to_ls_object(resp)
        assert isinstance(obj, Secret)
        assert obj.key == resp.get("data").get("key")
        assert obj.value == resp.get("data").get("value")
        assert obj.description == resp.get("data").get("description")

    def test_read_special_variable(self):
        access_key_id_test = "access_key_id_test"
        secret_access_key_test = "secret_access_key_test"
        access_key_id_default = "THE_ID_DEFAULT"
        secret_access_key_default = "THE_SECRET_DEFAULT"
        params = {
            "access_key_id": access_key_id_test,
            "secret_access_key": secret_access_key_test
        }
        access_key_id = util.read_special_variable(params, "access_key_id", default_value=None)
        secret_access_key = util.read_special_variable(params, "secret_access_key", default_value=None)
        assert access_key_id == access_key_id_test
        assert secret_access_key == secret_access_key_test

        params = {
            "access_key_id": access_key_id_test,
            "secret_access_key": secret_access_key_test
        }
        access_key_id = util.read_special_variable(params, "access_key_id", access_key_id_default)
        secret_access_key = util.read_special_variable(params, "secret_access_key", secret_access_key_default)
        assert access_key_id == access_key_id_default
        assert secret_access_key == secret_access_key_default

    def test_encode_datetime(self):
        utc_ts = util.encode_datetime(datetime(2023, 1, 1))
        assert isinstance(utc_ts, int) or isinstance(utc_ts, float)
        assert utc_ts <= 1672506000 + 60

    def test_api_encode(self):
        data = {"search": "live_search_text", "page": "1", "size": 10}
        encoded_params = list(util.api_encode(data))
        for encoded_param in encoded_params:
            assert encoded_param[1] == data.get(encoded_param[0])

    def test_merge_dicts(self):
        x = {"x_key": "x_value"}
        y = {"y_key": "y_value"}
        z_expected = x.copy()
        z_expected.update(y)
        z = util.merge_dicts(x, y)
        assert z == z_expected



import json
import os
import sys

from dotenv import load_dotenv
from locker.binary_adapter import BinaryAdapter
from locker.error import ResourceNotFoundError


class TestBinaryAdapter(object):
    @staticmethod
    def init_binary_adapter():
        load_dotenv()
        access_key_id = os.getenv('ACCESS_KEY_ID')
        secret_access_key = os.getenv('SECRET_ACCESS_KEY')
        binary_adapter = BinaryAdapter(
            api_base="https://api.locker.io/locker_secrets",
            access_key_id=access_key_id, secret_access_key=secret_access_key
        )
        return binary_adapter

    def test_get_platform(self):
        binary_adapter = self.init_binary_adapter()
        platform = binary_adapter.get_platform()
        assert platform == sys.platform

    def test_get_sdk_version(self):
        binary_adapter = self.init_binary_adapter()
        sdk_version = binary_adapter.get_sdk_version()
        assert isinstance(sdk_version, str) and len(sdk_version.split(".")) == 3

    def test_call(self):
        binary_adapter = self.init_binary_adapter()
        cli = ["secret", "list"]
        secrets = binary_adapter.call(cli)
        assert isinstance(secrets, list)
        for secret in secrets:
            assert isinstance(secret, dict)
            assert isinstance(secret.get('key'), str)
            assert isinstance(secret.get('value'), str)
            assert isinstance(secret.get('description'), str) or isinstance(secret.get('description'), None)

    def test_interpret_response(self):
        binary_adapter = self.init_binary_adapter()
        error_body = {
            "object": "error",
            "error": "not_found",
            "message": "Not found the resource"
        }
        error_body_raw = json.dumps(error_body)
        try:
            exc = binary_adapter.interpret_response(error_body_raw)
        except ResourceNotFoundError:
            pass


from __future__ import absolute_import, division, print_function

from locker.ls_resources.abstract.api_resource import APIResource


class CreateableAPIResource(APIResource):
    @classmethod
    def create(cls, access_key_id=None, secret_access_key=None, api_base=None, api_version=None, **params):
        return cls._static_call(
            [f"{cls.class_cli()}", "create"],
            access_key_id,
            secret_access_key,
            api_base=api_base,
            api_version=api_version,
            params=params,
        )



from __future__ import absolute_import, division, print_function

from six.moves.urllib.parse import quote_plus

from locker.error import CliRunError
from locker.ls_resources.abstract import ListableAPIResource, CreateableAPIResource, UpdateableAPIResource, \
    DeletableAPIResource, DetailableAPIResource


class Environment(ListableAPIResource, DetailableAPIResource,
                  CreateableAPIResource, UpdateableAPIResource, DeletableAPIResource):

    OBJECT_NAME = "environment"

    @classmethod
    def get_environment(cls, name, access_key_id=None, secret_access_key=None, api_base=None, api_version=None, **params):
        """
        Get Environment object by name
        :param name:
        :param access_key_id:
        :param secret_access_key:
        :param api_base:
        :param api_version:
        :param params:
        :return:
        """
        base = cls.class_cli()
        cli_ = [f"{base}", "get", "--name", f"{name}"]
        instance = cls(None, access_key_id, secret_access_key, **params)
        try:
            instance._call_and_refresh(
                cli_, access_key_id=access_key_id, secret_access_key=secret_access_key,
                api_base=api_base, api_version=api_version, params=params
            )
        except CliRunError as e:
            # TODO: Change the return result when not found the environment
            return None
        return instance

    @classmethod
    def retrieve_environment(cls, name, access_key_id=None, secret_access_key=None, api_base=None, api_version=None,
                             **params):
        """
        Get Environment object by name
        :param name:
        :param access_key_id:
        :param secret_access_key:
        :param api_base:
        :param api_version:
        :param params:
        :return:
        """
        base = cls.class_cli()
        cli_ = [f"{base}", "get", "--name", f"{name}"]
        instance = cls(None, access_key_id, secret_access_key, **params)
        # try:
        instance._call_and_refresh(
            cli_, access_key_id=access_key_id, secret_access_key=secret_access_key,
            api_base=api_base, api_version=api_version, params=params
        )
        # except CliRunError as e:
        #     # TODO: Handle exception
        #     return None
        return instance

    @classmethod
    def modify(cls, **params):
        name = params.get("name")
        cli = [f"{cls.class_cli()}", "update", "--name", f"{quote_plus(name)}"]

        if "external_url" in params:
            cli += ["--new-url", f"{params.get('external_url')}"]
        # if "new_name" in params:
        #     cli += ["--new-name", f"{params.get('new_name')}"]
        # if "new_description" in params:
        #     cli += ["--new-description", f"{params.get('new_description')}"]

        return cls._static_call(cli, params=params)

    @classmethod
    def create(cls, access_key_id=None, secret_access_key=None, api_base=None, api_version=None, **params):
        cli = [f"{cls.class_cli()}", "create"]
        if "external_url" in params:
            cli += ["--url", f"{params.get('external_url')}"]
        return cls._static_call(
            cli,
            access_key_id,
            secret_access_key,
            api_base=api_base,
            api_version=api_version,
            params=params,
        )



from __future__ import absolute_import, division, print_function

from locker.ls_resources.abstract.api_resource import APIResource
from locker.ls_resources.abstract.singleton_api_resource import (
    SingletonAPIResource,
)
from locker.ls_resources.abstract.createable_api_resource import (
    CreateableAPIResource,
)
from locker.ls_resources.abstract.updateable_api_resource import (
    UpdateableAPIResource,
)
from locker.ls_resources.abstract.deletable_api_resource import (
    DeletableAPIResource,
)
from locker.ls_resources.abstract.listable_api_resource import (
    ListableAPIResource,
)

from locker.ls_resources.abstract.detailable_api_resource import (
    DetailableAPIResource,
)



from __future__ import absolute_import, division, print_function

import six
from six.moves.urllib.parse import quote_plus

from locker import util, binary_adapter, error, DEFAULT_TIMEOUT
from locker.ls_object import LSObject


class APIResource(LSObject):
    @classmethod
    def retrieve(cls, id, access_key_id=None, secret_access_key=None, **params):
        instance = cls(id, access_key_id, secret_access_key, **params)
        instance.refresh()
        return instance

    def refresh(self):
        return self._call_and_refresh(self.instance_cli())

    @classmethod
    def class_cli(cls):
        if cls == APIResource:
            raise NotImplementedError(
                "APIResource is an abstract class. You should perform "
                "actions on its subclasses (e.g. Environment, Secret)"
            )
        # The base cli command is object names of this class and the namespaces are separated in object name
        # with periods (.) and in CLI commands with (-), so replace the former with the latter
        base = cls.OBJECT_NAME.replace(".", "-")
        return base

        # # Namespaces are separated in object names with periods (.) and in URLs
        # # with forward slashes (/), so replace the former with the latter.
        # base = cls.OBJECT_NAME.replace(".", "/")
        # return "/v1/%ss" % (base,)

    def instance_cli(self):
        instance_id = self.get("id")

        if not isinstance(instance_id, six.string_types):
            raise error.InvalidRequestError(
                "Could not determine which URL to request: %s instance "
                "has invalid ID: %r, %s. ID should be of type `str` (or"
                " `unicode`)" % (type(self).__name__, instance_id, type(instance_id)),
                "id",
            )

        base = self.class_cli()
        extn = quote_plus(instance_id)
        # The -i param is `id` of the object
        return '%s get --id %s' % (base, extn)

    # The `method_` and `url_` arguments are suffixed with an underscore to
    # avoid conflicting with actual request parameters in `params`.
    def _call(
        self,
        cli_,
        access_key_id=None,
        secret_access_key=None,
        api_base=None,
        api_version=None,
        asjson=True,
        timeout=DEFAULT_TIMEOUT,
        output_format="json",
        params=None,
    ):
        obj = LSObject._call(
            self,
            cli_,
            access_key_id,
            secret_access_key,
            api_base,
            api_version,
            asjson,
            timeout,
            output_format,
            params,
        )

        if type(self) is type(obj):
            self.refresh_from(obj)
            return self
        else:
            return obj

    # The 'cli_' is suffixed with an underscore to avoid conflicting with request parameters in `params`
    def _call_and_refresh(
        self,
        cli,
        access_key_id=None,
        secret_access_key=None,
        api_base=None,
        api_version=None,
        asjson=True, timeout=DEFAULT_TIMEOUT, output_format="json", params=None
    ):
        obj = LSObject._call(
            self,
            cli, access_key_id, secret_access_key, api_base, api_version,
            asjson, timeout, output_format, params
        )
        self.refresh_from(obj)
        return self

    # The 'cli_' is suffixed with an underscore to avoid conflicting with request parameters in `params`
    @classmethod
    def _static_call(
        cls,
        cli_,
        access_key_id=None,
        secret_access_key=None,
        api_base=None,
        api_version=None,
        asjson=True,
        timeout=DEFAULT_TIMEOUT,
        output_format="json",
        params=None,
    ):
        params = None if params is None else params.copy()
        access_key_id = util.read_special_variable(params, "access_key_id", access_key_id)
        secret_access_key = util.read_special_variable(params, "secret_access_key", secret_access_key)
        api_base = util.read_special_variable(params, "api_base", api_base)
        api_version = util.read_special_variable(params, "api_version", api_version)
        headers = util.read_special_variable(params, "headers", None)
        skip_cli_lines = util.read_special_variable(params, "skip_cli_lines", 0)
        logger = util.read_special_variable(params, "logger", None) if params else None
        resttime = util.read_special_variable(params, "resttime", None)
        fetch = util.read_special_variable(params, "fetch", None)

        binary_executor = binary_adapter.BinaryAdapter(
            access_key_id=access_key_id,
            secret_access_key=secret_access_key,
            api_base=api_base,
            api_version=api_version,
            headers=headers,
            logger=logger,
            resttime=resttime,
            fetch=fetch
        )
        res_data = binary_executor.call(
            cli=cli_, params=params, asjson=asjson, timeout=timeout, skip_cli_lines=skip_cli_lines,
            output_format=output_format
        )
        return util.convert_to_ls_object(
            res_data, access_key_id, secret_access_key, api_base, api_version, params
        )




import copy
import hashlib
import json
import logging
import os
import platform
import socket
import stat
import sys
import tempfile
import threading
import traceback
from typing import Dict, Any
import requests

from locker.ls_resources import Secret, Environment


LOCKER_LOG = os.environ.get("LOCKER_LOG")


class Locker:
    DEFAULT_OPTIONS = {
        "access_key_id": None,
        "secret_access_key": None,
        "thread_locking": False,
        "api_base": "https://api.locker.io/locker_secrets",
        "api_version": "v1",
        "proxy": None,
        "log": "error",
        "max_network_retries": 0,
        "skip_cli_lines": 0,
        "headers": {
            "Cache-Control": "no-cache",
            "Content-Type": "application/json",
        },
    }

    def __init__(
        self,
        access_key_id: str = None,
        secret_access_key: str = None,
        api_base: str = None,
        api_version: str = None,
        proxy: Any = None,
        log: str = None,
        thread_locking: bool = False,
        max_network_retries: int = 0,
        resttime: int = 120,
        fetch: bool = False,
        options: Dict = None,
    ):
        # self.download_binary()
        if options is None:
            options = {}
        if api_base:
            options["api_base"] = api_base
        if api_version:
            options["api_version"] = api_version
        if proxy:
            options["proxy"] = proxy
        log = log or LOCKER_LOG
        if log:
            options["log"] = log
        assert resttime >= -1
        options["resttime"] = resttime
        options["fetch"] = fetch
        options["thread_locking"] = thread_locking

        if max_network_retries:
            options["max_network_retries"] = max_network_retries
        if access_key_id:
            options["access_key_id"] = access_key_id
        if secret_access_key:
            options["secret_access_key"] = secret_access_key

        self._options: dict[str, Any] = copy.deepcopy(Locker.DEFAULT_OPTIONS)

        # Set Headers
        if "headers" in options:
            headers = copy.copy(options["headers"])
        else:
            headers = {}

        self._options.update(options)
        self._options["headers"].update(headers)

        # Set Logger
        logger = self._set_stream_logger(level=self._options.get('log'))
        self._options["logger"] = logger

        # Rip off trailing slash since all urls depend on that
        assert isinstance(self._options["api_base"], str)
        if self._options["api_base"].endswith("/"):
            self._options["api_base"] = self._options["api_base"][:-1]

        # if access_key_basic_auth:
        #     self._create_access_key_basic_auth(*access_key_basic_auth)

        self.is_valid_binary = False

        # Locking
        self.lock = None
        if thread_locking:
            self.lock = threading.Lock()

    # ---- This method is DEPRECATED from 0.1.1b1 ------------------- #
    # def _create_access_key_basic_auth(self, access_key_id: str, secret_access_key: str):
    #     self._options["access_key"] = f"{access_key_id}:{secret_access_key}"

    @property
    def access_key_id(self):
        return self._options.get("access_key_id")

    @access_key_id.setter
    def access_key_id(self, access_key_id_value):
        self._options.update({"access_key_id": access_key_id_value})

    @property
    def secret_access_key(self):
        return self._options.get("secret_access_key")

    @secret_access_key.setter
    def secret_access_key(self, secret_access_key_value):
        self._options.update({"secret_access_key": secret_access_key_value})

    @property
    def api_base(self):
        return str(self._options.get("api_base"))

    @api_base.setter
    def api_base(self, api_base_value):
        self._options.update({"api_base": api_base_value})

    @property
    def api_version(self):
        return str(self._options.get("api_version"))

    @property
    def log(self):
        return self._options.get("log")

    @log.setter
    def log(self, log_value):
        self._options.update({"log": log_value})
        logger = self._set_stream_logger(level=log_value)
        self._options["logger"] = logger

    @property
    def thread_locking(self):
        return self._options.get("thread_locking")

    @thread_locking.setter
    def thread_locking(self, thread_locking_value):
        self._options.update({"thread_locking": thread_locking_value})
        if thread_locking_value is True:
            self.lock = threading.Lock()
        else:
            self.lock = None

    @staticmethod
    def _set_stream_logger(level, name='locker'):
        assert level in ["debug", "info", "warning", "error"], "The log level is not valid"
        level = level.upper()
        format_string = '%(asctime)s {hostname} %(levelname)s %(message)s'.format(**{'hostname': socket.gethostname()})

        logger = logging.getLogger(name)
        logger.setLevel(level)
        handler = logging.StreamHandler()
        handler.setLevel(level)
        formatter = logging.Formatter(format_string)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        return logger

    @property
    def resttime(self):
        return self._options.get("resttime")

    @resttime.setter
    def resttime(self, resttime_value):
        self._options.update({"resttime": resttime_value})

    @property
    def fetch(self):
        return self._options.get("fetch")

    @fetch.setter
    def fetch(self, fetch_value):
        self._options.update({"fetch": fetch_value})

    @property
    def skip_cli_lines(self):
        return self._options.get("skip_cli_lines")

    @property
    def headers(self):
        return self._options.get("headers")

    @headers.setter
    def headers(self, custom_headers):
        self._options.update({"headers": custom_headers})

    @property
    def max_network_retries(self):
        return self._options.get("max_network_retries")

    def _translate_options(self, params):
        _params = copy.deepcopy(self._options)
        _params.update(params)
        return _params

    @staticmethod
    def __get_download_paths():
        _about_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), "__about__.json")
        with open(_about_file, 'r') as fd:
            about_data = json.load(fd)
            binary_checksum_dict = about_data.get("binary_checksum")
            binary_version = about_data.get("binary_version")
        home_dir = os.path.expanduser("~")
        locker_dir = os.path.join(home_dir, ".locker")
        # Check if the .locker directory exists, and create it if not
        if not os.path.exists(locker_dir):
            try:
                os.makedirs(locker_dir)
            except (PermissionError, OSError):
                home_dir = tempfile.gettempdir()
                locker_dir = os.path.join(home_dir, ".locker")
                if not os.path.exists(locker_dir):
                    os.makedirs(locker_dir)
        binary_file_path = os.path.join(locker_dir, f"locker_binary-{binary_version}")
        if sys.platform == "darwin":
            if platform.processor() == "arm":
                checksum = binary_checksum_dict.get("mac-arm64")
                binary_url = f"https://s.locker.io/download/locker-cli-mac-arm64-{binary_version}"
            else:
                checksum = binary_checksum_dict.get("mac-x64")
                binary_url = f"https://s.locker.io/download/locker-cli-mac-x64-{binary_version}"
        elif sys.platform == "win32":
            # binary_version = "1.0.60"
            binary_url = f"https://s.locker.io/download/locker-cli-win-x64-{binary_version}.exe"
            binary_file_path = os.path.join(locker_dir, f"locker_binary-{binary_version}.exe")
            checksum = binary_checksum_dict.get("windows")
        else:
            p = platform.processor() or platform.machine()
            if p in ["arm", "aarch64"]:
                binary_url = f"https://s.locker.io/download/locker-cli-linux-arm64-{binary_version}"
                checksum = binary_checksum_dict.get("linux-arm64")
            else:
                binary_url = f"https://s.locker.io/download/locker-cli-linux-x64-{binary_version}"
                checksum = binary_checksum_dict.get("linux")

        return {
            "locker_dir": locker_dir,
            "binary_url": binary_url,
            "binary_file_path": binary_file_path,
            "checksum": checksum,
        }

    def check_binary(self):
        if not self.__check_binary_file():
            raise Exception("The binary CLI is not downloaded correctly. Please remove .locker folder and retry again")

    def __check_binary_file(self):

        p = self.__get_download_paths()
        binary_file_path = p.get("binary_file_path")
        binary_url = p.get("binary_url")
        checksum = p.get("checksum")

        download_retry = 1
        max_download_retries = 10
        if not self.is_valid_binary:
            print("[+] Starting checking binary")
            while download_retry <= max_download_retries:
                if not os.path.exists(binary_file_path):
                    is_valid_binary = self.__download_and_check(
                        url=binary_url, binary_file_path=binary_file_path, correct_checksum=checksum
                    )
                else:
                    downloaded_checksum = self.__get_downloaded_checksum(binary_file_path)
                    is_valid_binary = downloaded_checksum == checksum or self.__download_and_check(
                        url=binary_url, binary_file_path=binary_file_path, correct_checksum=checksum
                    )
                if not is_valid_binary:
                    download_retry += 1
                    continue
                self.is_valid_binary = is_valid_binary
                break
        return self.is_valid_binary

    @staticmethod
    def __get_downloaded_checksum(binary_file_path):
        try:
            with open(binary_file_path, "rb") as f:
                binary_bytes = f.read()
                return hashlib.sha256(binary_bytes).hexdigest()
        except FileNotFoundError:
            return None

    def __download_and_check(self, url, binary_file_path, correct_checksum):
        downloaded = self.__download_binary_cli(url=url, save_path=binary_file_path)
        if not downloaded:
            return False
        return self.__get_downloaded_checksum(binary_file_path) == correct_checksum

    @staticmethod
    def __download_binary_cli(url, save_path):
        r = requests.get(url, stream=True)
        if r.ok:
            with open(save_path, 'wb') as f:
                for chunk in r.iter_content(chunk_size=1024 * 8):
                    if chunk:
                        f.write(chunk)
                        f.flush()
                        os.fsync(f.fileno())
            try:
                # Make the binary executable
                st = os.stat(save_path)
                os.chmod(save_path, st.st_mode | stat.S_IEXEC)
            except PermissionError as e:
                tb = traceback.format_exc()
                print(f"set permission error {e} - {tb}")
                pass
            return True

        # HTTP status code 4XX/5XX
        else:
            print("Download failed: status code {}\n{}".format(r.status_code, r.text))
            return False

    def list(self, **params):
        if self.thread_locking is True:
            with self.lock:
                self.check_binary()
                return Secret.list(**self._translate_options(params))
        else:
            self.check_binary()
            return Secret.list(**self._translate_options(params))

    def export(self, environment_name=None, output_format="dotenv",  **params):
        if self.thread_locking is True:
            with self.lock:
                self.check_binary()
                return Secret.export(
                    environment_name=environment_name, output_format=output_format, **self._translate_options(params)
                )
        else:
            self.check_binary()
            return Secret.export(
                environment_name=environment_name, output_format=output_format, **self._translate_options(params)
            )

    def get(self, key, environment_name=None, default_value=None, **params):
        if self.thread_locking is True:
            with self.lock:
                self.check_binary()
                return Secret.get_secret(
                    key,
                    environment_name=environment_name,
                    default_value=default_value,
                    **self._translate_options(params)
                )
        else:
            self.check_binary()
            return Secret.get_secret(
                key,
                environment_name=environment_name,
                default_value=default_value,
                **self._translate_options(params)
            )

    def get_secret(self, key, environment_name=None, default_value=None, **params):
        if self.thread_locking is True:
            with self.lock:
                self.check_binary()
                return Secret.get_secret(
                    key,
                    environment_name=environment_name,
                    default_value=default_value,
                    **self._translate_options(params)
                )
        else:
            self.check_binary()
            return Secret.get_secret(
                key,
                environment_name=environment_name,
                default_value=default_value,
                **self._translate_options(params)
            )

    def retrieve(self, key, environment_name=None, **params):
        if self.thread_locking is True:
            with self.lock:
                self.check_binary()
                return Secret.retrieve_secret(key, environment_name=environment_name, **self._translate_options(params))
        else:
            self.check_binary()
            return Secret.retrieve_secret(key, environment_name=environment_name, **self._translate_options(params))

    def create(self, **params):
        if self.thread_locking is True:
            with self.lock:
                self.check_binary()
                return Secret.create(**self._translate_options(params))
        else:
            self.check_binary()
            return Secret.create(**self._translate_options(params))

    def modify(self, **params):
        if self.thread_locking is True:
            with self.lock:
                self.check_binary()
                return Secret.modify(**self._translate_options(params))
        else:
            self.check_binary()
            return Secret.modify(**self._translate_options(params))

    def list_environments(self, **params):
        if self.thread_locking is True:
            with self.lock:
                self.check_binary()
                return Environment.list(**self._translate_options(params))
        else:
            self.check_binary()
            return Environment.list(**self._translate_options(params))

    def get_environment(self, name, **params):
        if self.thread_locking is True:
            with self.lock:
                self.check_binary()
                return Environment.get_environment(
                    name=name,
                    **self._translate_options(params)
                )
        else:
            self.check_binary()
            return Environment.get_environment(
                name=name,
                **self._translate_options(params)
            )

    def retrieve_environment(self, name, **params):
        if self.thread_locking is True:
            with self.lock:
                self.check_binary()
                return Environment.retrieve_environment(
                    name=name,
                    **self._translate_options(params)
                )
        else:
            self.check_binary()
            return Environment.retrieve_environment(
                name=name,
                **self._translate_options(params)
            )

    def create_environment(self, **params):
        if self.thread_locking is True:
            with self.lock:
                self.check_binary()
                return Environment.create(**self._translate_options(params))
        else:
            self.check_binary()
            return Environment.create(**self._translate_options(params))

    def modify_environment(self, **params):
        if self.thread_locking is True:
            with self.lock:
                self.check_binary()
                return Environment.modify(**self._translate_options(params))
        else:
            self.check_binary()
            return Environment.modify(**self._translate_options(params))



import logging
import os
import socket
import traceback


LOCKER_LOG = os.environ.get("LOCKER_LOG")


class Logger:
    def __init__(self, log_level=None):
        format_string = '%(asctime)s {hostname} %(levelname)s %(message)s'.format(**{'hostname': self._get_hostname()})
        format_log = logging.Formatter(format_string)

        self._log_level = self._set_level_log(log_level=log_level)
        # logging.basicConfig(level=self._log_level)
        self.locker_logger = logging.getLogger("locker")
        for handler in self.locker_logger.handlers:
            self.locker_logger.removeHandler(handler)

        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(format_log)
        self.locker_logger.addHandler(stream_handler)

    @staticmethod
    def _get_hostname():
        return socket.gethostname()

    @staticmethod
    def _set_level_log(log_level):
        _map_level_log = {
            "debug": logging.DEBUG,
            "info": logging.INFO,
            "warning": logging.WARNING,
            "error": logging.ERROR
        }
        if log_level in _map_level_log.keys():
            return _map_level_log[log_level]
        elif LOCKER_LOG in _map_level_log.keys():
            return _map_level_log[LOCKER_LOG]
        else:
            return logging.NOTSET

    def debug(self, msg):
        self.locker_logger.setLevel(self._log_level)
        self.locker_logger.debug(msg)

    def info(self, msg):
        self.locker_logger.setLevel(self._log_level)
        self.locker_logger.info(msg)

    def warning(self, msg):
        self.locker_logger.setLevel(self._log_level)
        self.locker_logger.warning(msg)

    def error(self, trace=None):
        self.locker_logger.setLevel(self._log_level)
        if trace is None:
            tb = traceback.format_exc()
            trace = 'Something was wrong' if tb is None else tb
        self.locker_logger.error(trace)



import json
import sys

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


with open('locker/__about__.json', 'r') as fd:
    _about_data = json.load(fd)
    __version__ = _about_data.get("version")
    binary_version = _about_data.get("binary_version")


def _requirements():
    # download_binary()
    with open('requirements.txt', 'r') as f:
        return [name.strip() for name in f.readlines()]


def _requirements_test():
    # download_binary()
    with open('requirements-test.txt', 'r') as f:
        return [name.strip() for name in f.readlines()]


def _classifiers():
    classifiers = [
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.14",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.6",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ]
    if "b1" in __version__ or "b2" in __version__:
        classifiers += ["Development Status :: 3 - Alpha"]
    else:
        classifiers += ["Development Status :: 5 - Production/Stable"]
    return classifiers


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)


with open('README.md', 'r') as f:
    long_description = f.read()


def main():
    setup(
        name="lockersm",
        version=__version__,
        author="CyStack",
        author_email="contact@locker.io",
        url="https://locker.io",
        download_url="",
        description="Locker Secret Python SDK",
        long_description=long_description,
        long_description_content_type="text/markdown",
        keywords=[
            "django",
            "vault management",
            "security"
        ],
        # license = BSD-3-Clause  # Example license
        include_package_data=True,
        packages=find_packages(
            exclude=[
                "docs",
                "examples",
                "tests",
                "tests.*",
                "venv",
                "projectenv",
                "*.sqlite3"
            ]
        ),
        python_requires=">=3.6",
        install_requires=_requirements(),
        tests_require=_requirements_test(),
        cmdclass={
            'test': PyTest,
        },
        classifiers=_classifiers(),
    )


if __name__ == "__main__":
    main()



from __future__ import absolute_import, division, print_function

from locker.ls_resources.abstract.api_resource import APIResource


class SingletonAPIResource(APIResource):
    @classmethod
    def retrieve(cls, **params):
        return super().retrieve(None, **params)

    @classmethod
    def class_cli(cls):
        if cls == SingletonAPIResource:
            raise NotImplementedError(
                "SingletonAPIResource is an abstract class. You should perform actions on its subclasses (e.g. Secre)"
            )
        # Namespaces are separated in object names with periods (.) and in URLs
        # with forward slashes (/), so replace the former with the latter.
        base = cls.OBJECT_NAME.replace(".", "/")
        return "/v1/%s" % (base,)

    def instance_cli(self):
        return self.class_cli()



# -*- coding: utf-8 -*-


"""Meta data of locker-sdk-python"""


__version__ = '1.0.5'
__author__ = 'CyStack Corporation'
__copyright__ = 'Copyright 2023, CyStack Corporation'
__license__ = ''

__all__ = (
    '__version__'
)



from __future__ import absolute_import, division, print_function

import hashlib
import json
import os
import platform
import stat
import sys
import tempfile
import traceback
import requests

# from locker.atomic_locking import AtomicNameLock


ROOT_PATH = os.path.dirname(os.path.realpath(__file__))

DEFAULT_TIMEOUT = 180

DEV_MODE = False

_about_file = os.path.join(ROOT_PATH, "__about__.json")
with open(_about_file, 'r') as fd:
    about_data = json.load(fd)
    binary_version = about_data.get("binary_version")
    binary_checksum_dict = about_data.get("binary_checksum")


home_dir = os.path.expanduser("~")
locker_dir = os.path.join(home_dir, ".locker")
# Check if the .locker directory exists, and create it if not
if not os.path.exists(locker_dir):
    try:
        os.makedirs(locker_dir)
    except (PermissionError, OSError):
        home_dir = tempfile.gettempdir()
        locker_dir = os.path.join(home_dir, ".locker")
        if not os.path.exists(locker_dir):
            os.makedirs(locker_dir)

binary_file_path = os.path.join(locker_dir, f"locker_binary-{binary_version}")
downloaded_checksum_path = os.path.join(locker_dir, f"CHECKSUM")

# Check os and get the binary url
if sys.platform == "darwin":
    if platform.processor() == "arm":
        binary_url = f"https://s.locker.io/download/locker-cli-mac-arm64-{binary_version}"
        checksum = binary_checksum_dict.get("mac-arm64")
    else:
        binary_url = f"https://s.locker.io/download/locker-cli-mac-x64-{binary_version}"
        checksum = binary_checksum_dict.get("mac-x64")
elif sys.platform == "win32":
    binary_url = f"https://s.locker.io/download/locker-cli-win-x64-{binary_version}.exe"
    binary_file_path = os.path.join(locker_dir, f"locker_binary-{binary_version}.exe")
    checksum = binary_checksum_dict.get("windows")
else:
    p = platform.processor() or platform.machine()
    if p in ["arm", "aarch64"]:
        binary_url = f"https://s.locker.io/download/locker-cli-linux-arm64-{binary_version}"
        checksum = binary_checksum_dict.get("linux-arm64")
    else:
        binary_url = f"https://s.locker.io/download/locker-cli-linux-x64-{binary_version}"
        checksum = binary_checksum_dict.get("linux")


# lock = AtomicNameLock('locker_downloader')
# if lock.acquire(timeout=30):
#     if not os.path.exists(binary_file_path):
#         r = requests.get(binary_url, stream=True)
#         if r.ok:
#             print("saving to", os.path.abspath(binary_file_path))
#             logging.debug(f"saving to {os.path.abspath(binary_file_path)}")
#             with open(binary_file_path, 'wb') as f:
#                 for chunk in r.iter_content(chunk_size=1024 * 8):
#                     if chunk:
#                         f.write(chunk)
#                         f.flush()
#                         os.fsync(f.fileno())
#             logging.warning(f"saving ok {os.path.abspath(binary_file_path)}")
#             try:
#                 # Make the binary executable
#                 # logging.warning(f"starting set permission {binary_file_path}")
#                 st = os.stat(binary_file_path)
#                 os.chmod(binary_file_path, st.st_mode | stat.S_IEXEC)
#                 # logging.warning(f"set permission ok {binary_file_path}")
#             except PermissionError as e:
#                 tb = traceback.format_exc()
#                 logging.error(f"set permission error {e} - {tb}")
#                 pass
#
#         # HTTP status code 4XX/5XX
#         else:
#             logging.error("Download failed: status code {}\n{}".format(r.status_code, r.text))
#             print("Download failed: status code {}\n{}".format(r.status_code, r.text))
#     lock.release()


def download_binary_cli(url, save_path):
    r = requests.get(url, stream=True)
    if r.ok:
        print("saving to", os.path.abspath(binary_file_path))
        # logging.debug(f"saving to {os.path.abspath(save_path)}")
        with open(save_path, 'wb') as f:
            for chunk in r.iter_content(chunk_size=1024 * 8):
                if chunk:
                    f.write(chunk)
                    f.flush()
                    os.fsync(f.fileno())

        print(f"saving ok {os.path.abspath(save_path)}")
        # logging.debug(f"saving ok {os.path.abspath(save_path)}")
        try:
            # Make the binary executable
            st = os.stat(save_path)
            os.chmod(save_path, st.st_mode | stat.S_IEXEC)
        except PermissionError as e:
            tb = traceback.format_exc()
            print(f"set permission error {e} - {tb}")
            # logging.error(f"set permission error {e} - {tb}")
            pass

        # Calc and save checksum
        with open(save_path, "rb") as f:
            binary_bytes = f.read()
            readable_hash = hashlib.sha256(binary_bytes).hexdigest()
        with open(downloaded_checksum_path, "w") as f:
            f.write(readable_hash)

        return True

    # HTTP status code 4XX/5XX
    else:
        # logging.error("Download failed: status code {}\n{}".format(r.status_code, r.text))
        print("Download failed: status code {}\n{}".format(r.status_code, r.text))
        return False


def get_downloaded_checksum(save_path):
    try:
        with open(downloaded_checksum_path, "r") as f:
            downloaded_checksum = f.read()
            return downloaded_checksum
    except FileNotFoundError:
        with open(save_path, "rb") as f:
            binary_bytes = f.read()
            readable_hash = hashlib.sha256(binary_bytes).hexdigest()
        with open(downloaded_checksum_path, "w") as f:
            f.write(readable_hash)
        return readable_hash


def download_and_check(url, save_path):
    downloaded = download_binary_cli(url=url, save_path=save_path)
    if not downloaded:
        return False
    return get_downloaded_checksum(binary_file_path) == checksum


download_retry = 1
max_download_retries = 10
# Currently, not check the checksum
is_valid_binary = True

if not is_valid_binary:
    while download_retry <= max_download_retries:
        if not os.path.exists(binary_file_path):
            is_valid_binary = download_and_check(url=binary_url, save_path=binary_file_path)
        else:
            downloaded_checksum = get_downloaded_checksum(binary_file_path)
            is_valid_binary = downloaded_checksum == checksum or download_and_check(
                url=binary_url, save_path=binary_file_path
            )
        if not is_valid_binary:
            download_retry += 1
            continue
        break

else:
    if not os.path.exists(binary_file_path):
        download_binary_cli(url=binary_url, save_path=binary_file_path)

if not is_valid_binary:
    raise Exception("The binary CLI is not downloaded correctly")


from locker.client import Locker




from __future__ import absolute_import, division, print_function

import json
import logging
import os
import stat
import sys
import subprocess

from six.moves.urllib.parse import urlencode

from locker import ROOT_PATH, DEV_MODE, DEFAULT_TIMEOUT, util, error, binary_file_path


class BinaryAdapter(object):
    def __init__(self, access_key_id=None, secret_access_key=None, api_base=None, api_version=None, headers=None,
                 logger=None, resttime: int = 120, fetch: bool = False):
        self.access_key_id = access_key_id
        self.secret_access_key = secret_access_key
        self.api_base = api_base
        self.api_version = api_version
        self.headers = headers
        self.system_platform = self.get_platform()
        self.logger = logger or logging.getLogger("locker")
        self.resttime = resttime
        self.fetch = fetch

    @classmethod
    def make_executable(cls, path):
        st = os.stat(path)
        os.chmod(path, st.st_mode | stat.S_IEXEC)

    @staticmethod
    def get_platform():
        # Return darwin/win32/linux
        return sys.platform

    @staticmethod
    def get_sdk_version():
        _about_file = os.path.join(ROOT_PATH, "__about__.json")
        with open(_about_file, 'r') as fd:
            version = json.load(fd).get("version")
        return version

    def get_binary_file(self):
        if DEV_MODE:
            # Checking the os system, returns the corresponding binary
            # OS X
            if self.system_platform == "darwin":
                return os.path.join(ROOT_PATH, "bin", "locker_mac")
            # Windows
            elif self.system_platform == "win32":
                return os.path.join(ROOT_PATH, "bin", "locker_windows.exe")
            # Default is linux
            else:
                return os.path.join(ROOT_PATH, "bin", "locker_linux")
        return binary_file_path

    def call(
        self,
        cli,
        params=None,
        asjson=True,
        timeout=DEFAULT_TIMEOUT,
        skip_cli_lines=0,
        output_format="json"
    ):
        # with self._lock:
        binary_file = self.get_binary_file()
        my_access_key_id = self.access_key_id
        my_secret_access_key = self.secret_access_key
        # if my_access_key_id is None or my_secret_access_key is None:
        #     raise error.AuthenticationError(
        #         "No Access key provided. (HINT: set your API key using "
        #         '"locker.access_key_id = <ACCESS-KEY-ID>" and "locker.secret_access_key = <SECRET-ACCESS-KEY>"). '
        #         "You can generate Access Key from the Locker Secret web interface."
        #     )
        my_headers = self.headers or {}
        default_client_agent = f"Python - {self.get_sdk_version()}"

        if isinstance(cli, list):
            cli_arr = cli
        else:
            self.logger.warning("[!] The string command will be deprecated in the feature")
            cli_arr = cli.split(' ')
        command_arr = [f'{binary_file}'] + cli_arr
        if my_access_key_id is not None:
            command_arr += [
                '--access-key-id', f'{my_access_key_id}'
            ]
        if my_secret_access_key is not None:
            command_arr += [
                '--secret-access-key', f'{my_secret_access_key}'
            ]

        command_arr += [
            '--api-base', f'{self.api_base}',
            '--agent', f'{default_client_agent}',
            '--resttime', f'{self.resttime}',
            # '--verbose'
        ]
        if output_format:
            command_arr += [f"--output-format", f"{output_format}"]
        if self.fetch is True:
            command_arr += ['--fetch']

        if my_headers:
            if isinstance(my_headers, dict):
                my_headers_list = [f"{k}:{v}" for k, v in my_headers.items()]
                my_headers = ",".join(my_headers_list)
            command_arr += ['--headers', f'{my_headers}']
        else:
            command_arr += ['--headers', '']

        # Building full command with params
        post_data = None
        if "get" in cli or "delete" in cli:
            encoded_params = urlencode(list(util.api_encode(params or {})))
            # Don't use strict form encoding by changing the square bracket control
            # characters back to their literals. This is fine by the server, and
            # makes these parameter strings easier to read.
            encoded_params = encoded_params.replace("%5B", "[").replace("%5D", "]")
            # TODO: Build api url by passing filter params to command
            # if params:
            #     abs_url = _build_api_url(abs_url, encoded_params)
            pass
        elif "update" in cli or "create" in cli:
            params_data = params or {}
            for k, v in params_data.items():
                if k in ["key", "value", "description", "new_key", "new_value", "new_description", "env",
                         "name", "url", "new_name", "new_url", "new_environment",
                         "new-key", "new-value", "new-environment", "new-description", "new-name", "new-url"]:
                    command_arr += [f'--{k.replace("_", "-")}', f'{v}']

        #     post_data = json.dumps(params or {})
        # if post_data:
        #     command_arr += ['--data', f'{post_data}']

        self.logger.debug(f"[+] Running with timeout {timeout} cli command: {command_arr}")
        try:
            raw = subprocess.check_output(
                command_arr,
                stderr=subprocess.STDOUT, shell=False, universal_newlines=True, timeout=timeout
            )
        except subprocess.TimeoutExpired as e:
            exc = error.CliRunError(e.stdout)
            exc.process = e
            raise exc
        except subprocess.CalledProcessError as e:
            signs = ['"success": false', '"success": true', '"object": "error"']
            if any(s in e.output for s in signs):
                raw = e.output
            elif str(e.output).strip() == 'Killed' or 'returned non-zero exit status 1' in str(e):
                exc = error.CliRunError(e.stdout)
                exc.process = e
                raise exc
            else:
                self.logger.warning(f"[!] subprocess.CalledProcessError: {e} {e.output}. The command is: {command_arr}")
                exc = error.CliRunError(e.stdout)
                exc.process = e
                raise exc
        return self.interpret_response(res_body=raw, asjson=asjson, skip_cli_lines=skip_cli_lines)

    def interpret_response(self, res_body, asjson=True, skip_cli_lines=0):
        # Skip cli lines
        if skip_cli_lines > 0:
            res_body = res_body.split("\n", skip_cli_lines)[skip_cli_lines]
        # Log break
        try:
            res_body = res_body.split("----------- LOG BREAK -----------")[1]
        except IndexError:
            pass
        if not asjson:
            return res_body
        try:
            if hasattr(res_body, "decode"):
                res_body = res_body.decode("utf-8")
        except Exception:
            self.logger.warning(f"[!] Invalid decode response body from CLI:::{res_body}")
            exc = error.CliRunError(
                f"Invalid decode response body from CLI: {res_body}",
                res_body
            )
            exc.process = res_body
            raise exc
        try:
            res_body = json.loads(res_body)
        except json.decoder.JSONDecodeError:
            self.logger.warning(f"[!] CLI result json decode error:::{res_body}")
            exc = error.CliRunError(
                f"CLI JSONDecodeError:::{res_body}", res_body
            )
            exc.process = res_body
            raise exc
        if self._should_handle_as_error(res_body):
            res_body.update({"object": "error"})
            self.handle_error_response(res_body)
        return res_body

    @staticmethod
    def _should_handle_as_error(res_body):
        try:
            return res_body.get("object") == "error" or res_body.get("success") is False or\
                res_body.get("success") == "false"
        except AttributeError:
            return False

    def handle_error_response(self, res_body):
        exc = self.specific_cli_error(error_data=res_body)
        self.logger.warning(f"[!] CLI return error object:::{exc.http_body}")
        raise exc

    def specific_cli_error(self, error_data):
        status_code = error_data.get("status_code")
        error_code = error_data.get("error")
        if status_code == 429 or error_code == "rate_limit":
            return error.RateLimitError(
                error_data.get("message"),
                http_body=error_data,
                http_status=429,
                code="rate_limit",
                status_code=429
            )
        elif status_code == 401 or error_code in ["unauthorized", "invalid_secret_access_key"]:
            return error.AuthenticationError(
                error_data.get("message"),
                http_body=error_data,
                http_status=401,
                code="unauthorized",
                status_code=401
            )
        elif status_code == 403 or error_code == "permission_denied":
            return error.PermissionDeniedError(
                error_data.get("message"),
                http_body=error_data,
                http_status=403,
                code="permission_denied",
                status_code=403
            )
        elif status_code == 404 or error_code in ["not_found", "not_found_error"]:
            return error.ResourceNotFoundError(
                error_data.get("message"),
                http_body=error_data,
                http_status=404,
                code="not_found",
                status_code=404
            )
        elif (status_code and status_code >= 500) or error_code == "server_error":
            return error.APIServerError(
                error_data.get("message"),
                http_body=error_data,
                http_status=500,
                code="server_error",
                status_code=500
            )
        elif error_code == "http_error":
            return error.APIConnectionError(
                error_data.get("message"),
                http_body=error_data,
                http_status=503,
                code="http_error",
                status_code=503
            )
        elif error_code in ["database_error", "command_error"]:
            return error.CliRunError(
                error_data.get("message"),
                http_body=error_data,
                code="cli_error",
            )
        else:
            return error.LockerError(
                error_data.get("message"),
                http_body=error_data,
                code="locker_error"
            )
